/* Distributed Checksum Clearinghouse
 *
 * reject messages contain URLs that resolve to DNS blacklisted IP addresses
 *
 * Copyright (c) 2008 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software without
 * changes for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear in all
 * copies and any distributed versions or copies are either unchanged
 * or not called anything similar to "DCC" or "Distributed Checksum
 * Clearinghouse".
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC by contacting Rhyolite Software
 * at sales@rhyolite.com.
 *
 * A commercial license would be for Distributed Checksum and Reputation
 * Clearinghouse software.  That software includes additional features.  This
 * free license for Distributed ChecksumClearinghouse Software does not in any
 * way grant permision to use Distributed Checksum and Reputation Clearinghouse
 * software
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.98-1.89 $Revision$
 */

#include "helper.h"
#include "dcc_heap_debug.h"
#ifndef DCC_WIN32
#include <sys/wait.h>
#include <arpa/inet.h>
#endif
#ifdef HAVE_RESOLV_H
#include <resolv.h>
#endif
#ifdef HAVE_ARPA_NAMESER_H
#include <arpa/nameser.h>
#endif

/* can check MX and NS addresses only with a standard resolver library */
#define MXNS_DNSBL
#if !defined(HAVE__RES) || !defined(HAVE_RES_INIT)
#undef HAVE__RES
#undef HAVE_RES_INIT
#undef MXNS_DNSBL
#endif
#if !defined (HAVE_RES_QUERY) || !defined(HAVE_DN_EXPAND)
#undef MXNS_DNSBL
#endif
#if !defined(HAVE_HSTRERROR)
#undef MXNS_DNSBL			/* MX lookups need raw hsterror() */
#endif
#if !defined(T_MX) || !defined(C_IN) || !defined(PACKETSZ)
#undef MXNS_DNSBL
#endif
#if !defined(RES_DEFNAMES) || !defined(RES_DNSRCH) || !defined(RES_NOALIASES)
#undef MXNS_DNSBL
#endif


DNSBL *dnsbls;

HELPER helper;
u_char have_helpers;
DCC_PATH dnsbl_progpath;

static u_char is_helper;

static u_char dnsbl_have_ipv6;		/* have DNSBL to check IPv6 addresses */
static u_char dnsbl_have_ipv4;		/* have DNSBL to check IPv4 addresses */
static DNSBL_FGS all_dnsbl_flags;

#define MAX_MSG_SECS 1000
#ifndef RES_TIMEOUT
#define RES_TIMEOUT 3
#endif
static int dnsbl_msg_secs = 25;		/* total seconds/mail message */
static time_t dnsbl_msg_us;
static int dnsbl_url_secs = 11;		/* total seconds/host name */
static time_t dnsbl_url_us;


static const char *
mx_ns_str(DNSBL_FGS flags)
{
	switch (flags) {
	case DNSBL_FG_CLIENT:
	case DNSBL_FG_MAIL_HOST:
	case DNSBL_FG_URL:
		return "";

	case DNSBL_FG_MX | DNSBL_FG_MAIL_HOST :
	case DNSBL_FG_MX | DNSBL_FG_URL:
		return " MX";

	case DNSBL_FG_NS | DNSBL_FG_MAIL_HOST:
	case DNSBL_FG_NS | DNSBL_FG_URL:
		return " NS";

	default:
		dcc_logbad(EX_SOFTWARE, "impossible DNSBL hit type");
		return "";
	}
}



/* make a string describing a DNSBL hit */
static u_char				/* 0=no hit */
hit2bt(DNSBL_WORK *blw)
{
	int i;

	switch (blw->hit_flags) {
	case DNSBL_FG_OFF:
		BUFCPY(blw->bt, "DNSBL checks off");
		return 0;
	case DNSBL_FG_NONE:
		BUFCPY(blw->bt, "no hit");
		return 0;
	case DNSBL_FG_CLIENT:
		BUFCPY(blw->bt, "SMTP client");
		dcc_su2str2(blw->result_str, ISZ(blw->result_str),
			    &blw->result);
		return 1;
	case DNSBL_FG_MAIL_HOST:
	case DNSBL_FG_MAIL_HOST | DNSBL_FG_MX:
	case DNSBL_FG_MAIL_HOST | DNSBL_FG_NS:
		i = snprintf(blw->bt, ISZ(blw->bt), "mail_host %s",
			     mx_ns_str(blw->hit_flags));
		if (i >= ISZ(blw->bt))
			strcpy(&blw->bt[ISZ(blw->bt)-4], "...");
		dcc_su2str2(blw->result_str, ISZ(blw->result_str),
			    &blw->result);
		return 1;
	case DNSBL_FG_URL:
	case DNSBL_FG_URL | DNSBL_FG_MX:
	case DNSBL_FG_URL | DNSBL_FG_NS:
		i = snprintf(blw->bt, ISZ(blw->bt), "URL%s %s",
			     mx_ns_str(blw->hit_flags), blw->tgt_dom);
		if (i >= ISZ(blw->bt))
			strcpy(&blw->bt[ISZ(blw->bt)-4], "...");
		dcc_su2str2(blw->result_str, ISZ(blw->result_str),
			    &blw->result);
		return 1;
	}

	dcc_logbad(EX_SOFTWARE, "impossible hit type");
	return 0;
}



/* Parse a string of the form:
 *	    "domain[,[IPaddr][,name|ipv4|ipv6]]
 *	or of one of the forms:
 *	    set:debug=X		more logging
 *	    set:msg_secs=S	total seconds checking blacklists/message
 *	    set:url_secs=S	total seconds per host name
 *	    set:[no-]envelope	obsolete, mail host & client IP address checks
 *	    set:[no-]client	client IP address checks
 *	    set:[no-]mail_host	envelope mail_from checks
 *	    set:[no-]URL	body URL checks
 *	    set:[no-]MX		MX checks
 *	    set:[no-]NS		NS checks
 *	    set:progpath	path of helper program
 *
 *	    set:helper=soc,fd,X	start DNS resolver process
 */
u_char					/* 0=bad */
dcc_parse_dnsbl(DCC_EMSG emsg, const char *entry, const char *progpath)
{
	static DNSBL_FGS cur_flags = (DNSBL_FG_CLIENT | DNSBL_FG_MAIL_HOST
				      | DNSBL_FG_URL | DNSBL_FG_MX
				      | DNSBL_FG_NS);
	static const void *cur_reply = 0;
	static int bl_num = 0;
	DNSBL *dp, *dp1, **dpp;
	const char *tgt_ip;		/* "hit" IP address of this blacklist */
	DNSBL_DOM tgt_ip_buf;
	enum DNSBL_TYPE bl_type;
	struct in6_addr tgt, tgt_mask;
	int tgt_bits;
	u_char tgt_use_ipv6;
	DCC_EMSG addr_emsg;
	int error, bl_dom_len, tgt_ip_addr_len;
	int val;
	char *p;
#ifdef HAVE_HELPERS
	int soc;
	int fd;
	int total_helpers;
#	define SAVE_ARG(arg) helper_save_arg("-B", arg)
#else
#	define SAVE_ARG(arg)
#endif

	if (progpath && dnsbl_progpath[0] == '\0') {
		const char *slash = strrchr(progpath, '/');
		if (slash)
			++slash;
		else
			slash = progpath;
		snprintf(dnsbl_progpath, sizeof(dnsbl_progpath),
			 "%.*sdns-helper", (int)(slash-progpath), progpath);
	}

	/* handle parameter settings */
	if (!CLITCMP(entry, "set:")) {
		const char *parm = entry+LITZ("set:");
#ifdef HAVE_HELPERS
		/* start running a helper process on the last, magic -B */
		if (3 == sscanf(parm, HELPER_PAT, &soc, &fd, &total_helpers)) {
			helper_child(soc, fd, total_helpers);
		}
#endif
		if (!CLITCMP(parm, "progpath=")) {
			BUFCPY(dnsbl_progpath, parm+LITZ("progpath="));
			return 1;
		}
		if (1 == sscanf(parm, "max_helpers=%d", &val)) {
			helper.max_helpers = val;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "debug")) {
			++helper.debug;
			SAVE_ARG(entry);
			return 1;
		}
		if (1 == sscanf(parm, "debug=%d", &val)) {
			helper.debug = val;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "envelope")) {	/* obsolete */
			cur_flags |= (DNSBL_FG_CLIENT | DNSBL_FG_MAIL_HOST);
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "no-envelope")) {	/* obsolete */
			cur_flags &= ~(DNSBL_FG_CLIENT | DNSBL_FG_MAIL_HOST);
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "client")) {
			/* obsolete */
			cur_flags |= DNSBL_FG_CLIENT;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "no-client")) {
			/* obsolete */
			cur_flags &= ~DNSBL_FG_CLIENT;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "mail_host")) {
			/* obsolete */
			cur_flags |= DNSBL_FG_MAIL_HOST;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "no-mail_host")) {
			/* obsolete */
			cur_flags &= ~DNSBL_FG_MAIL_HOST;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "body")) {    /* obsolete */
			cur_flags |= DNSBL_FG_URL;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "url")) {
			cur_flags |= DNSBL_FG_URL;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "no-body")) { /* obsolete */
			cur_flags &= ~DNSBL_FG_URL;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "no-URL")) {
			cur_flags &= ~DNSBL_FG_URL;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "mx")) {
#ifdef MXNS_DNSBL
			cur_flags |= DNSBL_FG_MX;
			SAVE_ARG(entry);
			return 1;
#else
			dcc_pemsg(EX_USAGE, emsg,
				  "MX DNS blacklist checks not supported");
			return 0;
#endif
		}
		if (!strcasecmp(parm, "no-mx")) {
			cur_flags &= ~DNSBL_FG_MX;
			SAVE_ARG(entry);
			return 1;
		}
		if (!strcasecmp(parm, "ns")) {
#ifdef MXNS_DNSBL
			cur_flags |= DNSBL_FG_MX;
			SAVE_ARG(entry);
			return 1;
#else
			dcc_pemsg(EX_USAGE, emsg,
				  "NS DNS blacklist checks not supported");
			return 0;
#endif
		}
		if (!strcasecmp(parm, "no-ns")) {
			cur_flags &= ~DNSBL_FG_NS;
			SAVE_ARG(entry);
			return 1;
		}
		if (!CLITCMP(parm, "rej-msg=")) {
			parm += LITZ("rej-msg=");
			if (*parm == '\0')
				cur_reply = 0;
			else
				cur_reply = dnsbl_parse_reply(parm);
			/* do not save for helpers */
			return 1;
		}
		if (!CLITCMP(parm, "msg-secs=")) {
			parm += LITZ("msg_secs=");
			val = strtoul(parm, &p, 10);
			if (*p != '\0' || val < 1 || val > MAX_MSG_SECS) {
				dcc_pemsg(EX_USAGE, emsg,
					  "bad number of seconds in \"-B %s\"",
					  entry);
				return 0;
			}
			if (dnsbl_msg_secs != val) {
				dnsbl_msg_secs = val;
				SAVE_ARG(entry);
			}
			return 1;
		}
		if (!CLITCMP(parm, "url-secs=")) {
			parm += LITZ("url_secs=");
			val = strtoul(parm, &p, 10);
			if (*p != '\0' || val < 1 || val > MAX_MSG_SECS) {
				dcc_pemsg(EX_USAGE, emsg,
					  "bad number of seconds in \"-B %s\"",
					  entry);
				return 0;
			}
			if (dnsbl_url_secs != val) {
				dnsbl_url_secs = val;
				SAVE_ARG(entry);
			}
			return 1;
		}
		dcc_pemsg(EX_USAGE, emsg, "unrecognized  \"-B %s\"",
			  entry);
		return 0;
	}

	/* we must have a DNSBL specification */
	bl_type = DNSBL_TYPE_IPV4;	/* assume it is a simple IPv4 DNSBL */
	tgt_ip = strchr(entry, ',');
	if (!tgt_ip) {
		bl_dom_len = strlen(entry);
		dnsbl_have_ipv4 = 1;
	} else {
		bl_dom_len = tgt_ip - entry;
		++tgt_ip;

		/* notice trailing ",name" or ",IPv4" */
		p = strchr(tgt_ip, ',');
		if (!p) {
			dnsbl_have_ipv4 = 1;
		} else {
			++p;
			if (!strcasecmp(p, "name")) {
				bl_type = DNSBL_TYPE_NAME;
			} else if (!strcasecmp(p, "IPV4")) {
				bl_type = DNSBL_TYPE_IPV4;
				dnsbl_have_ipv4 = 1;
			} else if (!strcasecmp(p, "IPV6")) {
				bl_type = DNSBL_TYPE_IPV6;
				dnsbl_have_ipv6 = 1;
			} else {
				dcc_pemsg(EX_NOHOST, emsg,
					  "unknown DNSBL type in \"%s\"",
					  entry);
				return 0;
			}
			STRLCPY(tgt_ip_buf, tgt_ip,
				min(ISZ(tgt_ip_buf), p-tgt_ip));
			tgt_ip = tgt_ip_buf;
		}
	}

	if (bl_dom_len < 1) {
		dcc_pemsg(EX_NOHOST, emsg,
			  "invalid DNS blacklist \"%s\"", entry);
		return 0;
	}

	/* assume 127.0.0.2 if the target address is missing */
	if (!tgt_ip || *tgt_ip == '\0')
		tgt_ip = "127.0.0.2";

	if (!strcasecmp(tgt_ip, "any")) {
		/* That we get a result when we lookup up an address in the
		 * in the DNSBL is all that matters in this case.
		 * Whether it is IPv6 or IPv4 does not */
		tgt_use_ipv6 = 2;
		tgt_ip_addr_len = INET6_ADDRSTRLEN;
		memset(&tgt, 0, sizeof(tgt));
		memset(&tgt_mask, 0, sizeof(tgt_mask));

	} else if (0 != (tgt_bits = dcc_str2cidr(emsg, &tgt, &tgt_mask,
						 &tgt_use_ipv6, tgt_ip,
						 0, 0))) {
		if (tgt_bits < 0) {
			dcc_pemsg(EX_NOHOST, addr_emsg,
				  "invalid DNS blacklist %s \"%s\"",
				  addr_emsg, entry);
			return 0;
		}
		tgt_ip_addr_len = INET6_ADDRSTRLEN;

	} else {
		dcc_host_lock();
		if (dcc_get_host(tgt_ip, 3, &error)) {
			/* prefer an IPv4 target address */
			if (dcc_hostaddrs[0].sa.sa_family == AF_INET) {
				tgt_use_ipv6 = 0;
				tgt_ip_addr_len = INET_ADDRSTRLEN;
				dcc_ipv4toipv6(&tgt,
					       dcc_hostaddrs[0].ipv4.sin_addr);
			} else {
				tgt_use_ipv6 = 1;
				tgt_ip_addr_len = INET6_ADDRSTRLEN;
				tgt = dcc_hostaddrs[0].ipv6.sin6_addr;
			}
			dcc_host_unlock();
			dcc_bits2mask(&tgt_mask, 128);

		} else {
			dcc_host_unlock();
			dcc_pemsg(EX_NOHOST, emsg,
				  "invalid DNSBL target IP address \"%s\": %s",
				  tgt_ip, DCC_HSTRERROR(error));
			return 0;
		}
	}

	if (bl_dom_len >= ISZ(DNSBL_DOM) - tgt_ip_addr_len) {
		dcc_host_unlock();
		/* we cannot fit the DNSBL base and the probe IP address
		 * into blw->probe.  We need to do DNS lookups of names
		 * like 33.22.11.10.dnsbl.example.com where the maximum length
		 * of "33.22.11.10" depends on tgt_ip_addr_len */
		dcc_pemsg(EX_NOHOST, emsg,
			  "DNSBL name \"%s\" too long", entry);
		return 0;
	}

	dp = dcc_malloc(sizeof(*dp));
	memset(dp, 0, sizeof(*dp));

	dp->tgt_use_ipv6 = tgt_use_ipv6;
	dp->tgt = tgt;
	dp->tgt_mask = tgt_mask;

	dp->bl_type = bl_type;
	dp->flags = cur_flags;
	dp->reply = cur_reply;
	all_dnsbl_flags |= cur_flags;
	memcpy(dp->bl_dom, entry, bl_dom_len);
	dp->bl_dom_len = bl_dom_len;

	dp->bl_num = ++bl_num;

	/* link the DNSBLs in the order they are declared so that they
	 * can be prioritized */
	dpp = &dnsbls;
	for (;;) {
		dp1 = *dpp;
		if (!dp1) {
			*dpp = dp;
			break;
		}
		/* notice sufficiently duplicate DNSBLS */
		if (!(dp->flags & DNSBL_FG_DUP) && !dp1->dup
		    && dp->tgt_use_ipv6 == dp1->tgt_use_ipv6
		    && dp->bl_type == dp1->bl_type
		    && dp->bl_dom_len == dp1->bl_dom_len
		    && !memcmp(dp->bl_dom, dp1->bl_dom, dp->bl_dom_len)) {
			dp->flags |= DNSBL_FG_DUP;
			dp1->dup = dp;
		}
		dpp = &dp1->fwd;
	}

	SAVE_ARG(entry);
	return 1;
#undef SAVE_ARG
}



/* resolve inconsistencies among the -B parameters */
static inline void
fix_url_secs(void)
{
	if (dnsbl_url_secs > dnsbl_msg_secs)
		dnsbl_url_secs =  dnsbl_msg_secs;
	dnsbl_msg_us = dnsbl_msg_secs * DCC_US;
	dnsbl_url_us = dnsbl_url_secs * DCC_US;
}



void
dcc_dnsbl_result(ASK_ST *ask_stp,
		 const REPLY_TPLT **reply, const DNSBL_WORK *blw)
{
	const DNSBL *dp;

	if (!blw)
		return;

	if (blw->timeouts)
		thr_log_print(blw->log_ctxt, 1,
			      "insufficient time for %d DNSBL checks\n",
			      blw->timeouts);

	if (blw->hit_flags == DNSBL_FG_NONE
	    || blw->hit_flags == DNSBL_FG_OFF)
		return;

	thr_log_print(blw->log_ctxt, 1, "DNSBL %s %s=%s\n",
		      blw->bt, blw->probe, blw->result_str);

	*ask_stp |= (ASK_ST_DNSBL_ISSPAM | ASK_ST_LOGIT);

	for (dp = dnsbls; dp; dp = dp->fwd) {
		if (blw->bl_num == dp->bl_num) {
			if (reply)
				*reply = dp->reply;
			return;
		}
	}
	dcc_logbad(EX_SOFTWARE, "failed to find responsible DNSBL");
}



/* start timer before we start to check DNS blacklists
 *	give up if it has already expired */
static u_char				/* 0=already too much time spent */
msg_secs_start(DNSBL_WORK *blw)
{
	if (blw->msg_us < 0) {
		++blw->timeouts;
		return 0;
	}

	if (blw->msg_us == 0) {
		/* out of time for the new domain before we start */
		blw->msg_us = -1;
		if (helper.debug)
			thr_trace_msg(blw->log_ctxt, "%s DNSBL"
				      " exhausted %d msg-secs before %s",
				      blw->id, dnsbl_msg_secs, blw->tgt_dom);
		else if (!is_helper)
			thr_log_print(blw->log_ctxt, 1, "%s DNSBL"
				      " exhausted %d msg-secs before %s\n",
				      blw->id, dnsbl_msg_secs, blw->tgt_dom);
		return 0;
	}

	gettimeofday(&blw->url_start, 0);
	blw->url_us = dnsbl_url_us;
	return 1;
}



static void
trace_secs(DNSBL_WORK *blw)
{
	if (is_helper) {
		if (helper.debug)
			thr_trace_msg(blw->log_ctxt,
				      "%s DNSBL helper exhausted "
				      "%d url-secs for %s",
				      blw->id, dnsbl_url_secs, blw->tgt_dom);
		return;
	}

	if (blw->msg_us > 0) {
		if (helper.debug)
			thr_trace_msg(blw->log_ctxt,
				      "%s DNSBL failed for %s,"
				      " %.1f msg-secs remaining",
				      blw->id, blw->tgt_dom,
				      blw->msg_us/(DCC_US*1.0));
		else
			thr_log_print(blw->log_ctxt, 1,
				      "%s DNSBL failed for %s,"
				      " %.1f msg-secs remaining\n",
				      blw->id, blw->tgt_dom,
				      blw->msg_us/(DCC_US*1.0));
	} else {
		if (helper.debug)
			thr_trace_msg(blw->log_ctxt,
				      "%s DNSBL exhausted %d msg-secs for %s",
				      blw->id, dnsbl_msg_secs, blw->tgt_dom);
		else
			thr_log_print(blw->log_ctxt, 1,
				      "%s DNSBL exhausted %d msg-secs for %s\n",
				      blw->id, dnsbl_msg_secs, blw->tgt_dom);
	}
}



/* see if we have run out of time */
static u_char				/* 0=timeout, 1=keep looking */
msg_secs_ck(DNSBL_WORK *blw)
{
	if (blw->msg_us < 0)
		return 0;

	gettimeofday(&blw->now, 0);
	blw->url_us_used = tv_diff2us(&blw->now, &blw->url_start);
	if (blw->url_us_used <= blw->url_us)
		return 1;

	blw->url_us = 0;
	blw->msg_us -= blw->url_us_used;
	if (helper.debug > 1
	    || (!is_helper && helper.debug > 0))
		trace_secs(blw);
	return 0;
}



/* account for time used */
static void
msg_secs_fin(DNSBL_WORK *blw,
	     u_char timeout)		/* 1=ran out of time */
{
	time_t used;

	gettimeofday(&blw->now, 0);
	used = tv_diff2us(&blw->now, &blw->url_start);
	if (used < 0)			/* handle clock jumps */
		return;

	blw->msg_us -= used;
	if (blw->msg_us <= 0) {
		if (!timeout) {
			/* ensure a log message on the next check, if any */
			blw->msg_us = 0;
			return;
		}
		blw->msg_us = -1;
	}

	if (blw->msg_us < 0
	    || timeout)
		trace_secs(blw);
}



#ifndef HAVE__RES
static void
dnsbl_res_delays(const DNSBL_WORK *blw UATTRIB)
{
	return;
}
#else
/* Limit resolver delays to as much as we are willing to wait
 *	We should be talking to a local caching resolver.  If it does not answer
 *	immediately, it is unlikely to later.  If it does eventually get an
 *	answer, the answer will probably ready the next time we ask.
 *  dcc_host_lock() must be held. */
static void
dnsbl_res_delays(const DNSBL_WORK *blw)
{
	int budget;			/* seconds we can afford */
	int res_retrans;		/* _res.retrans, retransmition delay */
	int res_retry;			/* _res.retry, # of retransmissions */
	int total;			/* retrans*retry = worst case delay */

	/* get the current value */
	if (!_res.options & RES_INIT)
		res_init();
	res_retry = _res.retry;
	res_retrans = _res.retrans;
	if (!res_retry)
		res_retry = 4;
	if (!res_retrans)
		res_retrans = RES_TIMEOUT;

	/* assume binary exponential backoffs as in the BIND resolver */
	total = ((1<<res_retry) -1) * res_retrans;

	/* If the default values could take too long, then try 2 seconds.
	 * If that is still too long, go to 1 retransmission and an initial
	 * retransmission delay of 1/3 of the total allowed delay.
	 * 1/3 from one exponential backoff for a total of 2**2-1=3 times
	 * the initial delay.
	 * We should be using a local caching DNS server and so should not
	 * see many lost packets */
	budget = (blw->url_us - blw->url_us_used + DCC_US/2) /DCC_US;
	if (total >= budget) {
		res_retry = 2;		/* 2 retries are often few enough */
		total = ((1<<res_retry) -1) * res_retrans;

		/* if that is not few enough,
		 * then reduce the retransmission delay to fit */
		if (total >= budget) {
			res_retrans = budget/3;
			if (res_retrans == 0)
				res_retrans = 1;
		}
	}

	if (_res.retry != res_retry
	    || _res.retrans != res_retrans) {
		if (helper.debug > 4)
			thr_trace_msg(blw->log_ctxt,
				      "DNSBL change _res.retry from %d to %d,"
				      " _res.retrans from %d to %d seconds",
				      _res.retry, res_retry,
				      _res.retrans, res_retrans);
		_res.retry = res_retry;
		_res.retrans = res_retrans;
	}
}
#endif /* !HAVE__RES */



/* get ready to handle a mail message */
void
dcc_dnsbl_init(DCC_GOT_CKS *cks,
	       DCC_CLNT_CTXT *dcc_ctxt, CMN_WORK *log_ctxt, const char *id)
{
	DNSBL_WORK *blw;

	if (!dnsbls)
		return;

	blw = cks->dnsbl;
	if (!blw) {
		blw = dcc_malloc(sizeof(*cks->dnsbl));
		memset(blw, 0, sizeof(*cks->dnsbl));
		cks->dnsbl = blw;

		/* general initializations on the first use of DNS blacklists */
		fix_url_secs();
	}

	blw->bl_num = 0x7fffffff;
	blw->hit_flags = DNSBL_FG_NONE;
	blw->id = id;
	blw->tgt_dom[blw->tgt_dom_len = 0] = '\0';
	blw->timeouts = 0;
	blw->msg_us = dnsbl_msg_us;

	blw->dcc_ctxt = dcc_ctxt;
	blw->log_ctxt = log_ctxt;
}



/* look for a host name or IP address in a DNS blacklist.
 *	These DNS operations should be done with local default values for
 *	RES_DEFNAMES, RES_DNSRCH, and RES_NOALIASES because the blacklist
 *	might be something local and strange. */
static DNSBL_BL_RESULT			/* -1=no time, 0=miss, or hit DNSBL # */
lookup(DNSBL_WORK *blw,
       const char *probe,		/* check this name */
       const DNSBL *dp,			/* start with this blacklist */
       DNSBL_FGS flags)			/* type of lookup */
{
	char sustr[DCC_SU2STR_SIZE];
	const DCC_SOCKU *hap;
	struct in6_addr addr6;
	const struct in6_addr *addr6p;
	int error;

	if (!msg_secs_ck(blw))
		return -1;

	/* get the list of addresses for the probe in the DNSBL */
	dcc_host_lock();
	dnsbl_res_delays(blw);
	if (!dcc_get_host(probe, dp->tgt_use_ipv6, &error)) {
		if (!msg_secs_ck(blw))
			return -1;
		if (helper.debug > 1)
			thr_trace_msg(blw->log_ctxt,
				      "%s DNSBL%s %s gethostbyname(%s): %s",
				      blw->id, mx_ns_str(flags),
				      blw->tgt_dom, probe,
				      DCC_HSTRERROR(error));
		dcc_host_unlock();
		return 0;
	}

	/* check each address obtained for the probe in the DNSBL
	 * for a hit in the list of duplicate references to this DNSBLs */
	do {
		if ((dp->flags & flags) != flags)
			continue;
		for (hap = dcc_hostaddrs; hap < dcc_hostaddrs_end; ++hap) {
			/* finished if any answer is good enough
			 * or if we get a match */
			if (hap->sa.sa_family == AF_INET6) {
				addr6p = &hap->ipv6.sin6_addr;
			} else {
				addr6p = &addr6;
				dcc_ipv4toipv6(&addr6, hap->ipv4.sin_addr);
			}
			if (DCC_IN_BLOCK(*addr6p, dp->tgt, dp->tgt_mask)) {
				if (helper.debug > 1)
					thr_trace_msg(blw->log_ctxt,
						      "%s DNSBL%s hit %s"
						      " gethostbyname(%s)=%s",
						      blw->id, mx_ns_str(flags),
						      blw->tgt_dom, probe,
						      dcc_su2str2(sustr,
							  sizeof(sustr),
							  hap));
				blw->result = *hap;
				dcc_host_unlock();
				/* got a hit, so return the blacklist ordinal */
				return dp->bl_num;
			}
		}

		/* check the results in other references to the blacklist */
	} while ((dp = dp->dup) != 0);

	/* no hit */
	if (helper.debug > 1) {
		thr_trace_msg(blw->log_ctxt,
			      "%s DNSBL%s miss %s gethostbyname(%s)=%s%s",
			      blw->id, mx_ns_str(flags), blw->tgt_dom, probe,
			      dcc_su2str2(sustr, sizeof(sustr),
					  &dcc_hostaddrs[0]),
			      dcc_hostaddrs_end == dcc_hostaddrs ? "" : " ...");
	}
	dcc_host_unlock();
	return 0;
}



/* check one IPv4 address against the DNS blacklists */
static DNSBL_BL_RESULT			/* -1=no time, 0=miss, or hit DNSBL # */
dnsbl_ipv4(DNSBL_WORK *blw,
	   char *probe, int probe_len,
	   const struct in_addr *ipv4,
	   DNSBL_FGS flags)		/* type of lookup */
{
	const DNSBL *dp;
	const u_char *bp;
	int ret;

	bp = (u_char *)ipv4;
	for (dp = dnsbls; dp; dp = dp->fwd) {
		if (dp->bl_type != DNSBL_TYPE_IPV4)
			continue;
		if ((dp->flags & flags) != flags)
			continue;
		if (dp->flags & DNSBL_FG_DUP)
			continue;
		snprintf(probe, probe_len, "%d.%d.%d.%d.%s",
			 bp[3], bp[2], bp[1], bp[0], dp->bl_dom);
		ret = lookup(blw, probe, dp, flags);
		if (ret) {
			if (ret > 0) {
				blw->hit_flags = flags;
				dcc_ipv4tostr(blw->bip, sizeof(blw->bip), ipv4);
				hit2bt(blw);
			}
			return ret;
		}
	}
	return 0;
}



/* check one IPv6 address against the DNS blacklists */
static DNSBL_BL_RESULT			/* -1=no time, 0=miss, or hit DNSBL # */
dnsbl_ipv6(DNSBL_WORK *blw,
	   char *probe, int probe_len,
	   const struct in6_addr *ipv6,
	   DNSBL_FGS flags)
{
	const DNSBL *dp;
	const u_char *bp;
	int ret;

	bp = (u_char *)ipv6;
	for (dp = dnsbls; dp; dp = dp->fwd) {
		if (dp->bl_type != DNSBL_TYPE_IPV6)
			continue;
		if ((dp->flags & flags) != flags)
			continue;
		if (dp->flags & DNSBL_FG_DUP)
			continue;
		snprintf(probe, probe_len,
			 "%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%s",
			 bp[15], bp[14], bp[13], bp[12],
			 bp[11], bp[10], bp[9], bp[8],
			 bp[7], bp[6], bp[5], bp[4],
			 bp[3], bp[2], bp[1], bp[0],
			 dp->bl_dom);
		ret = lookup(blw, probe, dp, flags);
		if (ret) {
			if (ret > 0) {
				blw->hit_flags = flags;
				dcc_ipv6tostr(blw->bip, sizeof(blw->bip), ipv6);
				hit2bt(blw);
			}
			return ret;
		}
	}
	return 0;
}



/* convert a name to one or more IP addresses to be checked in a DNS blacklist.
 *  dcc_host_lock() must be held.
 *	These DNS operations need RES_DEFNAMES and RES_DNSRCH off and
 *	RES_NOALIASES on when the name is an MX or NS server name.
 */
static u_char				/* 0=failed */
dnsbl_get_host(const DNSBL_WORK *blw,
	       const char *dom, u_char use_ipv6,
	       int *errorp,		/* put error number here */
	       DNSBL_FGS flags		/* 0, DNSBL_FG_MX, or DNSBL_FG_NS */
#ifndef MXNS_DNSBL
	       UATTRIB
#endif
	       )
{
#ifdef MXNS_DNSBL
	u_long save_options;
#endif
	u_char result;

#ifdef MXNS_DNSBL
	save_options = _res.options;
	if (flags & (DNSBL_FG_MX | DNSBL_FG_NS)) {
		_res.options &= ~(RES_DEFNAMES | RES_DNSRCH);
		_res.options |= RES_NOALIASES;
	}
#endif
	dnsbl_res_delays(blw);
	result = dcc_get_host(dom, use_ipv6, errorp);
#ifdef MXNS_DNSBL
	if (flags & (DNSBL_FG_MX | DNSBL_FG_NS))
		_res.options = save_options;
#endif
	return result;
}



/* look for a domain name in the DNS blacklists */
static DNSBL_BL_RESULT			/* -1=no time, 0=miss, or hit DNSBL # */
dnsbl_name_sub(DNSBL_WORK *blw,
	       char *probe, int probe_len,  /* put DNS blacklist probe here */
	       const char *dom,		/* see if this domain is blacklisted */
	       DNSBL_FGS flags)		/* type of lookup */
{
	const DNSBL *dp;
	const DCC_SOCKU *sup;
	struct in_addr ipv4[8];
	struct in6_addr ipv6[4];
	int ret, i, error;

	/* check the name in any host name DNSBLs that we have */
	for (dp = dnsbls; dp; dp = dp->fwd) {
		if (dp->bl_type != DNSBL_TYPE_NAME)
			continue;
		if ((dp->flags & flags) != flags)
			continue;
		if (dp->flags & DNSBL_FG_DUP)
			continue;
		/* truncate on the left */
		i = probe_len-1-1-dp->bl_dom_len;
		if (i > 0) {
			i = strlen(dom) - i;
			if (i < 0)
				i = 0;
			snprintf(probe, probe_len, "%s.%s",
				 dom+i, dp->bl_dom);
			ret = lookup(blw, probe, dp, flags);
			if (ret) {
				if (ret > 0) {
					blw->hit_flags = flags;
					BUFCPY(blw->bip, dom);
					hit2bt(blw);
				}
				return ret;
			}
			if (!msg_secs_ck(blw))
				return -1;
		}
	}

	/* check IPv4 addresses for the name IP DNSBLs after the name itself */
	if (dnsbl_have_ipv4) {
		if (!msg_secs_ck(blw))
			return -1;
		dcc_host_lock();
		if (!dnsbl_get_host(blw, dom, 0, &error, flags)) {

			dcc_host_unlock();
			if (!msg_secs_ck(blw))
				return -1;
			if (helper.debug > 1)
				thr_trace_msg(blw->log_ctxt,
					      "%s DNSBL%s gethostbyname(%s):"
					      " %s",
					      blw->id, mx_ns_str(flags), dom,
					      DCC_HSTRERROR(error));
		} else {
			/* Try several of the IP addresses for the domain.
			 * Save any IP addresses we want to check before we
			 * check them, because checking changes the array
			 * of addresses. */
			for (sup = dcc_hostaddrs, i = 0;
			     sup < dcc_hostaddrs_end && i < DIM(ipv4);
			     ++sup, ++i) {
				ipv4[i] = sup->ipv4.sin_addr;
			}
			dcc_host_unlock();
			/* check the addresses in all of the DNS blacklists */
			do {
				ret = dnsbl_ipv4(blw, probe, probe_len,
						 &ipv4[--i], flags);
				if (ret)
					return ret;
			} while (i > 0);
		}
	}

	/* try some IPv6 addresses for the name if we have any IPv6 DNSBLs */
	if (dnsbl_have_ipv6) {
		if (!msg_secs_ck(blw))
			return -1;
		dcc_host_lock();
		if (!dnsbl_get_host(blw, dom, 1, &error, flags)) {
			dcc_host_unlock();
			if (!msg_secs_ck(blw))
				return -1;
			if (helper.debug > 1)
				thr_trace_msg(blw->log_ctxt,
					      "%s DNSBL%s gethostbyname(%s):"
					      " %s",
					      blw->id, mx_ns_str(flags), dom,
					      DCC_HSTRERROR(error));
		} else {
			for (sup = dcc_hostaddrs, i = 0;
			     sup < dcc_hostaddrs_end && i < DIM(ipv6);
			     ++sup, ++i) {
				ipv6[i] = sup->ipv6.sin6_addr;
			}
			dcc_host_unlock();
			do {
				ret = dnsbl_ipv6(blw, probe, probe_len,
						 &ipv6[--i], flags);
				if (ret)
					return ret;
			} while (i > 0);
		}
	}

	return 0;
}



static DNSBL_BL_RESULT			/* -1=no time, 0=miss, or hit DNSBL # */
dnsbl_addr(DNSBL_WORK *blw,
	   char *probe, int probe_len,	/* build DNS blacklist probe here */
	   const struct in6_addr *addr)
{
	struct in_addr ipv4;
	int ret;

	/* try IPv4 first */
	if (dnsbl_have_ipv4
	    && dcc_ipv6toipv4(&ipv4, addr)) {
		ret = dnsbl_ipv4(blw, probe, probe_len, &ipv4, DNSBL_FG_CLIENT);
		if (ret)
			return ret;
	}

	if (dnsbl_have_ipv6)
		return dnsbl_ipv6(blw, probe, probe_len, addr, DNSBL_FG_CLIENT);
	return 0;
}




#ifdef MXNS_DNSBL
static DNSBL_BL_RESULT			/* -2=NXDOMAIN, -1=time, 0, DNSBL # */
dnsbl_name_rr(DNSBL_WORK *blw,
	      char *probe, int probe_len,   /* put DNS blacklist probe here */
	      const char *dom,		/* see if this domain is blacklisted */
	      DNSBL_FGS flags)
{
	union {
	    u_char  buf[PACKETSZ+20];
	    HEADER  hdr;
	} answer;
	DNSBL_DOM responsible;
	u_char *ap, *eom;
	int cnt, skip, req_type, resp_type, ret, i;

	if (flags & DNSBL_FG_MX) {
		req_type = T_MX;
	} else if (flags & DNSBL_FG_NS) {
		req_type = T_NS;
	} else {
		dcc_logbad(EX_SOFTWARE, "impossible RR type");
	}

	dnsbl_res_delays(blw);
	i = res_query(dom, C_IN, req_type, answer.buf, sizeof(answer.buf));
	if (i < 0) {
		/* use raw hstrerror() here because we are using the
		 * raw resolver */
		if (helper.debug > 1)
			thr_trace_msg(blw->log_ctxt,
				      "%s DNSBL%s res_query(%s): %s",
				      blw->id, mx_ns_str(flags), dom,
				      hstrerror(h_errno));
		return -2;
	}

	ap = &answer.buf[HFIXEDSZ];
	if (i > ISZ(answer.buf))
		i = ISZ(answer.buf);
	eom = &answer.buf[i];

	/* skip the question */
	cnt = ntohs(answer.hdr.qdcount);
	while (--cnt >= 0) {
		skip = dn_skipname(ap, eom);
		if (skip < 0) {
			if (helper.debug > 1)
				thr_trace_msg(blw->log_ctxt,
					      "%s DNSBL%s dn_skipname(%s)=%d",
					      blw->id, mx_ns_str(flags),
					      dom, skip);
			return 0;
		}
		ap += skip+QFIXEDSZ;
	}

	cnt = ntohs(answer.hdr.ancount);
	while (--cnt >= 0 && ap < eom) {
		skip = dn_expand(answer.buf, eom, ap,
				 responsible, sizeof(responsible));
		if (skip < 0) {
			if (helper.debug > 1)
				thr_trace_msg(blw->log_ctxt,
					      "%s DNSBL%s dn_expand(%s)=%d",
					      blw->id, mx_ns_str(flags),
					      dom, skip);
			return 0;
		}

		ap += skip;
		GETSHORT(resp_type, ap);
		ap += 2+4;		/* skip class and TTL */
		GETSHORT(skip, ap);	/* get rdlength */
		if (resp_type != req_type) {
			ap += skip;
			continue;
		}

		if (req_type == T_MX)
			ap += 2;	/* skip MX preference */

		skip = dn_expand(answer.buf, eom, ap,
				 responsible, sizeof(responsible)-1);
		if (skip < 0) {
			if (helper.debug > 1)
				thr_trace_msg(blw->log_ctxt,
					      "%s DNSBL%s dn_expand(%s)=%d",
					      blw->id, mx_ns_str(flags),
					      dom, skip);
			return 0;
		}
		ap += skip;

		if (*responsible == '\0') {
			if (helper.debug > 1)
				thr_trace_msg(blw->log_ctxt,
					      "%s DNSBL%s null name for %s",
					      blw->id, mx_ns_str(flags), dom);
			continue;
		}

		ret = dnsbl_name_sub(blw, probe, probe_len, responsible, flags);
		if (ret)
			return ret;
	}

	return 0;
}



static DNSBL_BL_RESULT			/* -2=NXDOMAIN, -1=time, 0, DNSBL # */
dnsbl_name_ns(DNSBL_WORK *blw,
	      char *probe, int probe_len,   /* put DNS blacklist probe here */
	      const char *dom,		/* see if this domain is blacklisted */
	      DNSBL_FGS flags)
{
	const char *pdom;
	int dom_len;
	int level;
	int ret;

	dom_len = strlen(dom);
	if (dom_len == 0)
		return 0;

	pdom = dom + dom_len-1;

	/* check parent domains starting with the second level */
	level = 0;
	for (;;) {
		if (pdom == dom)
			return dnsbl_name_rr(blw, probe, probe_len,
					     dom, flags);

		if (*--pdom == '.'
		    && ++level >= 2) {
			ret = dnsbl_name_rr(blw, probe, probe_len,
					    pdom+1, flags);
			if (ret)
				return ret;
		}
	}
}
#endif /* MXNS_DNSBL */



/* look for a domain name in the DNS blacklists, including MX & NS servers */
static DNSBL_BL_RESULT			/* -1=no time, 0=miss, or hit DNSBL # */
dnsbl_name(DNSBL_WORK *blw,
	   char *probe, int probe_len,  /* put DNS blacklist probe here */
	   const char *dom,		/* see if this domain is blacklisted */
	   DNSBL_FGS flag)		/* DNSBL_FG_MAIL_HOST or DNSBL_FG_URL */
{
#ifdef MXNS_DNSBL
	u_char is_addr;
#ifndef NO_IPV6
	struct dcc_in6_addr sin6_addr;
#endif /* NO_IPV6 */
#endif /* MXNS_DNSBL */
	int result;
	DNSBL_FGS hit_flags;

	/* give up if none of the DNS blacklists allow this kind of search */
	if (!(all_dnsbl_flags & flag))
		return 0;

	result = 0;
#ifdef MXNS_DNSBL
	/* First try NS records because they are quickest
	 * but only if at least one DNS blacklist cares
	 * and if the address is a name and not numeric
	 * Start with the parent domains up to the second level
	 * and in the fastest order. */
	is_addr = (INADDR_NONE != inet_addr(dom));
#ifndef NO_IPV6
	if (!is_addr && 0 < inet_pton(AF_INET6, dom, &sin6_addr))
		is_addr = 1;
#endif
	hit_flags = flag | DNSBL_FG_NS;
	if (!is_addr
	    && (all_dnsbl_flags & hit_flags) == hit_flags) {
		result = dnsbl_name_ns(blw, probe, probe_len, dom, hit_flags);
		/* no NS RR means there are no other RRs */
		if (result == -2)
			return 0;
	}
#endif /* MXNS_DNSBL */

	if (!result) {
		hit_flags = flag;
		result = dnsbl_name_sub(blw, probe, probe_len, dom, hit_flags);
	}

#ifdef MXNS_DNSBL
	/* check MX servers if allowed by at least one DNS blacklist */
	if (!result && !is_addr) {
		hit_flags = flag | DNSBL_FG_MX;
		if ((all_dnsbl_flags & hit_flags) == hit_flags) {
			result = dnsbl_name_rr(blw, probe, probe_len, dom,
					       hit_flags);
			if (result == -2)
				result = 0; /* should not get NXDOMAIN now */
		}
	}
#endif /* MXNS_DNSBL */

	if (result > 0)
		blw->hit_flags = hit_flags;
	return result;
}



#ifdef HAVE_HELPERS
/* do some DNSBL work in a helper process */
u_char					/* 1=try to send the response */
dnsbl_work(const DNSBL_REQ *req, DNSBL_RESP *resp)
{
	DNSBL_WORK blw;
	DNSBL_BL_RESULT result;

	memset(&blw, 0, sizeof(blw));
	blw.url_start = req->hdr.start;
	blw.msg_us = MAX_MSG_SECS*DCC_US*2;
	blw.url_us = req->hdr.avail_us;
	BUFCPY(blw.tgt_dom, req->tgt_dom);
	blw.id = req->hdr.id;

	if (!is_helper) {
		/* this must be the first job for this helper process */
		is_helper = 1;
		fix_url_secs();
	}

	result = 0;
	switch (req->hit_flag) {
	case DNSBL_FG_CLIENT:		/* SMTP client IP address */
		result = dnsbl_addr(&blw, resp->probe, ISZ(resp->probe),
				    &req->tgt_addr);
		break;

	case DNSBL_FG_MAIL_HOST:	/* envelope mail_from */
		result = dnsbl_name(&blw, resp->probe, ISZ(resp->probe),
				    req->tgt_dom, DNSBL_FG_MAIL_HOST);
		break;

	case DNSBL_FG_URL:		/* URL in body */
		result = dnsbl_name(&blw, resp->probe, ISZ(resp->probe),
				    req->tgt_dom, DNSBL_FG_URL);
		break;

	default:
		dcc_logbad(EX_SOFTWARE, "%s DNSBL helper unknown type %d",
			   req->hdr.id, req->hit_flag);
	}

	if (result <= 0) {
		resp->bl_num = result;
		resp->hit_flags = DNSBL_FG_NONE;
	} else {
		resp->bl_num = result;
		resp->hit_flags = blw.hit_flags;
		resp->result = blw.result;
		BUFCPY(resp->bip, blw.bip);
	}

	return 1;
}



/* ask a helper process to check for something in the DNS blacklists */
static void
use_helper(DNSBL_WORK *blw,
	   DNSBL_FGS flag,		/* look for this in blw->tgt_dom */
	   const struct in6_addr *tgt_addr) /*	or this */
{
	DNSBL_REQ req;
	DNSBL_RESP resp;

	memset(&req, 0, sizeof(req));

	BUFCPY(req.hdr.id, blw->id);
	req.hit_flag = flag;
	switch (flag) {
	case DNSBL_FG_CLIENT:
		memset(req.tgt_dom, 0, sizeof(req.tgt_dom));
		memcpy(&req.tgt_addr, tgt_addr, sizeof(req.tgt_addr));
		break;
	case DNSBL_FG_MAIL_HOST:
		BUFCPY(req.tgt_dom, blw->tgt_dom);
		memset(&req.tgt_addr, 0, sizeof(req.tgt_addr));
		break;
	case DNSBL_FG_URL:
		BUFCPY(req.tgt_dom, blw->tgt_dom);
		memset(&req.tgt_addr, 0, sizeof(req.tgt_addr));
		break;
	default:
		dcc_logbad(EX_SOFTWARE, "unknown DNSBL type %d", flag);
	}

	BUFCPY(req.hdr.id, blw->id);
	BUFCPY(req.tgt_dom, blw->tgt_dom);
	if (tgt_addr)
		memcpy(&req.tgt_addr, tgt_addr, sizeof(req.tgt_addr));
	else
		memset(&req.tgt_addr, 0, sizeof(req.tgt_addr));

	if (!ask_helper(blw->dcc_ctxt, blw->log_ctxt, dnsbl_url_us,
			&req.hdr, sizeof(req), &resp.hdr, sizeof(resp))) {
		msg_secs_fin(blw, 1);
		return;
	}

	blw->bl_num = resp.bl_num;
	blw->hit_flags = resp.hit_flags;
	BUFCPY(blw->probe, resp.probe);
	blw->result = resp.result;
	BUFCPY(blw->bip, resp.bip);
	if ((hit2bt(blw) && helper.debug > 1)
	    || helper.debug > 3)
		thr_trace_msg(blw->log_ctxt, "%s DNSBL helper %s %s=%s",
			      blw->id, blw->bt,
			      blw->probe, blw->result_str);

	msg_secs_fin(blw, 0);
}
#endif /* HAVE_HELPERS */



void
dcc_dnsbl_url(DNSBL_WORK *blw)
{
	DNSBL_BL_RESULT result;

	/* nothing to do if no body URL DNS blacklists have been configured
	 * or if we already have a hit */
	if (blw->hit_flags != DNSBL_FG_NONE
	    || !(all_dnsbl_flags & DNSBL_FG_URL))
		return;

	/* or if we have already spent too much time checking blacklists */
	if (!msg_secs_start(blw))
		return;

#ifdef HAVE_HELPERS
	if (have_helpers) {
		use_helper(blw, DNSBL_FG_URL, 0);
		return;
	}
#endif

	result = dnsbl_name(blw, blw->probe, sizeof(blw->probe),
			    blw->tgt_dom, DNSBL_FG_URL);
	if (0 < result) {
		blw->bl_num = result;
		hit2bt(blw);
		if (helper.debug > 1)
			thr_trace_msg(blw->log_ctxt, "%s DNSBL %s %s=%s",
				      blw->id, blw->bt,
				      blw->probe, blw->result_str);
	}
	msg_secs_fin(blw, 0);
}



void
dcc_client_dnsbl(DNSBL_WORK *blw, const struct in6_addr *addr)
{
	DNSBL_BL_RESULT result;
	char *p;

	/* nothing to do if no DNS blacklists have been configured
	 * or if we already have a hit */
	if (!blw || blw->hit_flags != DNSBL_FG_NONE
	    || !(all_dnsbl_flags & DNSBL_FG_CLIENT))
		return;

	p = blw->tgt_dom;
	memcpy(p, "sender ", LITZ("sender "));
	p += LITZ("sender ");
	dcc_ipv6tostr2(p, sizeof(blw->tgt_dom)-LITZ("sender "), addr);

	/* or if we have already spent too much time checking blacklists */
	if (!msg_secs_start(blw))
		return;

#ifdef HAVE_HELPERS
	if (have_helpers) {
		use_helper(blw, DNSBL_FG_CLIENT, addr);
		return;
	}
#endif

	result = dnsbl_addr(blw, blw->probe, sizeof(blw->probe), addr);
	if (0 < result) {
		blw->bl_num = result;
		hit2bt(blw);
		if (helper.debug > 1)
			thr_trace_msg(blw->log_ctxt, "%s DNSBL %s %s=%s",
				      blw->id, blw->bt,
				      blw->probe, blw->result_str);
	}
	msg_secs_fin(blw, 0);
}



void
dcc_mail_host_dnsbl(DNSBL_WORK *blw, const char *host)
{
	DNSBL_BL_RESULT result;

	/* nothing to do if no DNS blacklists have been configured
	 * or if we already have a hit
	 * or if we have already spent too much time checking blacklists */
	if (!blw || blw->hit_flags != DNSBL_FG_NONE
	    || !(all_dnsbl_flags & DNSBL_FG_MAIL_HOST)
	    || !host || *host == '\0')
		return;

	/* or if we have already spent too much time checking blacklists */
	BUFCPY(blw->tgt_dom, host);
	if (!msg_secs_start(blw))
		return;

#ifdef HAVE_HELPERS
	if (have_helpers) {
		use_helper(blw, DNSBL_FG_MAIL_HOST, 0);
		return;
	}
#endif

	result = dnsbl_name(blw, blw->probe, sizeof(blw->probe),
			    host, DNSBL_FG_MAIL_HOST);
	if (0 < result) {
		blw->bl_num = result;
		hit2bt(blw);
		if (helper.debug > 1)
			thr_trace_msg(blw->log_ctxt, "%s DNSBL %s %s=%s",
				      blw->id, blw->bt,
				      blw->probe, blw->result_str);
	}
	msg_secs_fin(blw, 0);
}
