/* Distributed Checksum Clearinghouse
 *
 * combine RRD databases
 *
 * Copyright (c) 2008 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software without
 * changes for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice appear in all
 * copies and any distributed versions or copies are either unchanged
 * or not called anything similar to "DCC" or "Distributed Checksum
 * Clearinghouse".
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC by contacting Rhyolite Software
 * at sales@rhyolite.com.
 *
 * A commercial license would be for Distributed Checksum and Reputation
 * Clearinghouse software.  That software includes additional features.  This
 * free license for Distributed ChecksumClearinghouse Software does not in any
 * way grant permision to use Distributed Checksum and Reputation Clearinghouse
 * software
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.98-1.6 $Revision$
 */

#include "dcc_defs.h"
#include <math.h>

#define PROGNAME "rrd-combine: "

/* *********************************************************************** */
/* copied from rrd.h and rrd_format.h */
typedef double       rrd_value_t;

typedef union unival {
    unsigned long u_cnt;
    rrd_value_t   u_val;
} unival;

typedef struct stat_head_t {
    char	    cookie[4];
    char	    version[5];
    double	    float_cookie;
    unsigned long   ds_cnt;
    unsigned long   rra_cnt;
    unsigned long   pdp_step;
    unival	    par[10];
} stat_head_t;

#define DS_NAM_SIZE   20
#define DST_SIZE   20
typedef struct ds_def_t {
    char	    ds_nam[DS_NAM_SIZE];
    char	    dst[DST_SIZE];
    unival	    par[10];
} ds_def_t;

#define CF_NAM_SIZE   20
#define MAX_RRA_PAR_EN 10
typedef struct rra_def_t {
    char	    cf_nam[CF_NAM_SIZE];
    unsigned long   row_cnt;
    unsigned long   pdp_cnt;
    unival	    par[MAX_RRA_PAR_EN];
} rra_def_t;


typedef struct live_head_t {
    time_t	    last_up;		 /* when was rrd last updated */
    long	    last_up_usec;
} live_head_t;

#define LAST_DS_LEN 30
typedef struct pdp_prep_t{
    char	    last_ds[LAST_DS_LEN];
    unival	    scratch[10];
} pdp_prep_t;

#define MAX_CDP_PAR_EN 10
typedef struct cdp_prep_t{
    unival	    scratch[MAX_CDP_PAR_EN];
} cdp_prep_t;

typedef struct rra_ptr_t {
    unsigned long   cur_row;
} rra_ptr_t;

typedef struct rrd_t {
    struct stat	    sb;
    stat_head_t     *stat_head;		/* the static header */
    ds_def_t	    *ds_def;		/* list of data source definitions */
    rra_def_t	    *rra_def;		/* list of round robin archive def */
    live_head_t     *live_head;
    pdp_prep_t      *pdp_prep;		/* pdp data prep area */
    cdp_prep_t      *cdp_prep;		/* cdp prep area */
    rra_ptr_t	    *rra_ptr;		/* list of rra pointers */
    rrd_value_t     *rrd_value;		/* list of rrd values */
} rrd_t;

/* *********************************************************************** */


static DCC_EMSG dcc_emsg;

static const char *ofile;
static u_char force;
static int ofd = -1;
static rrd_t o_rrd;
static rrd_t i_rrd;


static void NRATTRIB
usage(void)
{
	fprintf(stderr, "usage: [-f] -o ofile ifile1 ifile2 ...\n");
	exit(1);
}


static void
map_rrd(rrd_t *rrdp)
{
	rrdp->ds_def = (ds_def_t *)&rrdp->stat_head[1];
	rrdp->rra_def = (rra_def_t *)&rrdp->ds_def[rrdp->stat_head->ds_cnt];
	rrdp->live_head =(live_head_t*)&rrdp->rra_def[rrdp->stat_head->rra_cnt];
	if (!strcmp(rrdp->stat_head->version, "0001")) {
		rrdp->pdp_prep = (pdp_prep_t *)(&rrdp->live_head->last_up + 1);
	} else {
		rrdp->pdp_prep = (pdp_prep_t *)(rrdp->live_head + 1);
	}
	rrdp->cdp_prep = (cdp_prep_t *)&rrdp->pdp_prep[rrdp->stat_head->ds_cnt];
	rrdp->rra_ptr = (rra_ptr_t *)&rrdp->cdp_prep[rrdp->stat_head->rra_cnt
						     * rrdp->stat_head->ds_cnt];
	rrdp->rrd_value =(rrd_value_t*)&rrdp->rra_ptr[rrdp->stat_head->rra_cnt];
}



static void
rrd_stat(rrd_t *rrdp, int fd, const char *fnm)
{
	if (0 > fstat(fd, &rrdp->sb)) {
		fprintf(stderr, PROGNAME"stat(%s): %s\n",
			fnm, strerror(errno));
		exit(1);
	}
	if (rrdp->sb.st_size <= ISZ(rrdp->stat_head)) {
		fprintf(stderr, PROGNAME"%s has too small size %d\n",
			fnm, (int)rrdp->sb.st_size);
		exit(1);
	}
}



static void
map_o_rrd(void)
{
	rrd_stat(&o_rrd, ofd, ofile);

	o_rrd.stat_head = (stat_head_t *)mmap(0, o_rrd.sb.st_size,
					      PROT_READ | PROT_WRITE,
					      MAP_SHARED,
					      ofd, 0);
	if (o_rrd.stat_head == (stat_head_t *)MAP_FAILED) {
		fprintf(stderr, PROGNAME"mmap(%s): %s\n",
			ofile, strerror(errno));
		exit(1);
	}

	map_rrd(&o_rrd);
}



static void
read_ifile(const char *fnm)
{
#	define PAD 128
	const rra_def_t *i_def, *o_def;
	int ifd, old_len, len, i;

	ifd = open(fnm, O_RDONLY, 0);
	if (ifd < 0) {
		fprintf(stderr, PROGNAME"open(%s): %s\n",
			fnm, strerror(errno));
		exit(1);
	}

	old_len = i_rrd.sb.st_size;
	rrd_stat(&i_rrd, ifd, fnm);

	if (old_len == 0) {
		i_rrd.stat_head = malloc(i_rrd.sb.st_size+PAD);

	} else if (i_rrd.sb.st_size >= old_len + PAD) {
		fprintf(stderr, PROGNAME"%s has %d bytes, more than %d\n",
			fnm, (int)i_rrd.sb.st_size, old_len);
		exit(1);
	}

	len = read(ifd, i_rrd.stat_head, i_rrd.sb.st_size);
	if (len != i_rrd.sb.st_size) {
		if (len < 0)
			fprintf(stderr, PROGNAME"read(%s,%d): %s\n",
				fnm, (int)i_rrd.sb.st_size, strerror(errno));
		else
			fprintf(stderr, PROGNAME"read(%s,%d)=%d\n",
				fnm, (int)i_rrd.sb.st_size, len);
		exit(1);
	}

	map_rrd(&i_rrd);

	if (old_len != 0) {
		if (i_rrd.stat_head->rra_cnt != o_rrd.stat_head->rra_cnt) {
			fprintf(stderr,
				PROGNAME"%ld instead of %ld RRAs in %s\n",
				i_rrd.stat_head->rra_cnt,
				o_rrd.stat_head->rra_cnt,
				fnm);
			exit(1);
		}
		if (i_rrd.stat_head->ds_cnt != o_rrd.stat_head->ds_cnt) {
			fprintf(stderr,
				PROGNAME"%ld instead of %ld DSs in %s\n",
				i_rrd.stat_head->ds_cnt,
				o_rrd.stat_head->ds_cnt,
				fnm);
			exit(1);
		}
		if (i_rrd.stat_head->pdp_step != o_rrd.stat_head->pdp_step) {
			fprintf(stderr,
				PROGNAME"%ld instead of %ld step in %s\n",
				i_rrd.stat_head->pdp_step,
				o_rrd.stat_head->pdp_step,
				fnm);
			exit(1);
		}
		for (i = 0, i_def = i_rrd.rra_def, o_def = o_rrd.rra_def;
		     i < (int)o_rrd.stat_head->rra_cnt;
		     ++i, ++i_def, ++o_def) {
			if (o_def->row_cnt != i_def->row_cnt) {
				fprintf(stderr,
					PROGNAME"%ld instead of %ld"
					" rows in RRA #%d in %s\n",
					i_def->row_cnt, o_def->row_cnt,
					i, fnm);
				exit(1);
			}
			if (o_def->pdp_cnt != i_def->pdp_cnt) {
				fprintf(stderr,
					PROGNAME"%ld instead of %ld"
					" data points in RRA #%d in %s\n",
					i_def->pdp_cnt, o_def->pdp_cnt,
					i, fnm);
				exit(1);
			}
		}
	}

	close(ifd);
}



int NRATTRIB
main(int argc, char **argv)
{
	int fno;
	int len;
	rrd_value_t *i_base, *o_base;
	int rows, cols, rrd, row, i_row, o_row, col;
	rrd_value_t i_val, o_val;
	struct timeval tv;
	struct stat sb;
	time_t newest_mtime;
	int newest;
	int i;

	while ((i = getopt(argc, argv, "fo:")) != -1) {
		switch (i) {
		case 'f':
			force = 1;
			break;
		case 'o':
			ofile = optarg;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 2 || !ofile)
		usage();

	/* find the newest file */
	newest_mtime = 0;
	newest = 0;
	for (fno = 0; fno < argc; ++fno) {
		if (0 > stat(argv[fno], &sb)) {
			fprintf(stderr, PROGNAME"stat(%s): %s\n",
				argv[fno], strerror(errno));
			exit(1);
		}
		if (newest_mtime < sb.st_mtime) {
			newest_mtime = sb.st_mtime;
			newest = fno;
		}
	}

	/* create and mmap() the output file */
	ofd = open(ofile,
		   O_RDWR | O_CREAT | (force ? O_TRUNC : O_EXCL),
		   0666);
	if (ofd < 0) {
		fprintf(stderr, PROGNAME"open(%s): %s\n",
			ofile, strerror(errno));
		exit(1);
	}

	/* copy the newest input file to the output file */
	read_ifile(argv[newest]);
	len = write(ofd, i_rrd.stat_head, i_rrd.sb.st_size);
	if (len != (int)i_rrd.sb.st_size) {
		fprintf(stderr, PROGNAME"write(%s, %d) = %d: %s\n",
			ofile, (int)i_rrd.sb.st_size, len, strerror(errno));
		exit(1);
	}
	map_o_rrd();

	for (fno = 0; fno < argc; ++fno) {
		if (fno == newest)
			continue;

		read_ifile(argv[fno]);

		i_base = i_rrd.rrd_value;
		o_base = o_rrd.rrd_value;
		for (rrd = 0; rrd < (int)o_rrd.stat_head->rra_cnt; ++rrd) {
			rows = o_rrd.rra_def[rrd].row_cnt;
			cols = i_rrd.stat_head->ds_cnt;

			/* find last row in the two RRDs numbered as
			 * data consolidation moments since the UNIX epoch */
			i_row = (i_rrd.live_head->last_up
				 / (i_rrd.rra_def[rrd].pdp_cnt
				    * i_rrd.stat_head->pdp_step));

			o_row = (o_rrd.live_head->last_up
				 / (o_rrd.rra_def[rrd].pdp_cnt
				    * o_rrd.stat_head->pdp_step));

			/* Find the number of rows to combine. */
			i = o_row - i_row;
			if (i >= 0) {
				/* If the output RRD is newer than the input,
				 * then we will add only some of the input
				 * rows. */
				row = rows - i;
			} else {
				/* If the output is older,
				 * then we have problems. */
				fprintf(stderr, PROGNAME"%s newer than %s\n",
					argv[fno], argv[1]);
				exit(1);
			}

			i_row = (i_rrd.rra_ptr[rrd].cur_row + 1) * cols;
			o_row = (o_rrd.rra_ptr[rrd].cur_row + 1) * cols;
			do {
				/* wrap to the start at the last row */
				if (i_row >= rows*cols)
					i_row = 0;
				if (o_row >= rows*cols)
					o_row = 0;

				for (col = 0;
				     col < cols;
				     ++col, ++i_row, ++o_row) {
					i_val = i_base[i_row];
					if (isnan(i_val))
					    continue;
					o_val = o_base[o_row];
					if (isnan(o_val)) {
					    o_val = i_val;
					} else {
					    o_val += i_val;
					}
					o_base[o_row] = o_val;
				}
			} while (--row > 0);

			i_base += rows * cols;
			o_base += rows * cols;
		}
	}


	tv.tv_usec = 0;
	tv.tv_sec = o_rrd.live_head->last_up;
	fsync(ofd);
	if (!dcc_set_mtime(dcc_emsg, ofile, ofd, &tv)) {
		fprintf(stderr, PROGNAME"%s", dcc_emsg);
		exit(1);
	}

	exit(0);
}
